#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import random
from unittest import mock

from oslo_config import cfg
from oslo_db.sqlalchemy import utils as db_utils
from oslo_utils import uuidutils
import sqlalchemy as sa

from ironic.common import context
from ironic.common import exception
from ironic.common import release_mappings
from ironic.common import states
from ironic.db import api as db_api
from ironic.tests.unit.db import base
from ironic.tests.unit.db import utils


CONF = cfg.CONF


class UpgradingTestCase(base.DbTestCase):

    def setUp(self):
        super(UpgradingTestCase, self).setUp()
        self.dbapi = db_api.get_instance()
        self.object_versions = release_mappings.get_object_versions()

    def test_check_versions_emptyDB(self):
        # nothing in the DB
        self.assertTrue(self.dbapi.check_versions())

    @mock.patch.object(db_utils, 'column_exists', autospec=True)
    def test_check_versions_missing_version_columns(self, column_exists):
        column_exists.return_value = False
        self.assertRaises(exception.DatabaseVersionTooOld,
                          self.dbapi.check_versions)

    @mock.patch.object(release_mappings, 'get_object_versions', autospec=True)
    @mock.patch.object(db_utils, 'column_exists', autospec=True)
    def test_check_versions_handles_missing_table(
            self, column_exists, mock_release_mappings):
        column_exists.side_effect = sa.exc.NoSuchTableError('meow')
        mock_release_mappings.return_value = {'Node': {'1.0'}}
        self.assertTrue(
            self.dbapi.check_versions(permit_initial_version=True))
        self.assertEqual(1, column_exists.call_count)

    @mock.patch.object(release_mappings, 'get_object_versions', autospec=True)
    @mock.patch.object(db_utils, 'column_exists', autospec=True)
    def test_check_versions_raises_missing_table(
            self, column_exists, mock_release_mappings):
        column_exists.side_effect = sa.exc.NoSuchTableError('meow')
        mock_release_mappings.return_value = {'Node': {'1.0', '1.1'}}
        self.assertRaises(sa.exc.NoSuchTableError, self.dbapi.check_versions)
        self.assertEqual(1, column_exists.call_count)

    def test_check_versions(self):
        for v in self.object_versions['Node']:
            node = utils.create_test_node(uuid=uuidutils.generate_uuid(),
                                          version=v)
            node = self.dbapi.get_node_by_id(node.id)
            self.assertEqual(v, node.version)
        self.assertTrue(self.dbapi.check_versions())

    def test_check_versions_node_no_version(self):
        node = utils.create_test_node(version=None)
        node = self.dbapi.get_node_by_id(node.id)
        self.assertIsNone(node.version)
        self.assertFalse(self.dbapi.check_versions())

    def test_check_versions_ignore_node(self):
        node = utils.create_test_node(version=None)
        node = self.dbapi.get_node_by_id(node.id)
        self.assertIsNone(node.version)
        self.assertTrue(self.dbapi.check_versions(ignore_models=['Node']))

    def test_check_versions_node_old(self):
        node = utils.create_test_node(version='1.0')
        node = self.dbapi.get_node_by_id(node.id)
        self.assertEqual('1.0', node.version)
        self.assertFalse(self.dbapi.check_versions())

    def test_check_versions_conductor(self):
        for v in self.object_versions['Conductor']:
            # NOTE(jroll) conductor model doesn't have a uuid :(
            conductor = utils.create_test_conductor(
                hostname=uuidutils.generate_uuid(), version=v,
                id=random.randint(1, 1000000))
            conductor = self.dbapi.get_conductor(conductor.hostname)
            self.assertEqual(v, conductor.version)
        self.assertTrue(self.dbapi.check_versions())

    def test_check_versions_conductor_old(self):
        conductor = utils.create_test_conductor(version='1.0')
        conductor = self.dbapi.get_conductor(conductor.hostname)
        self.assertEqual('1.0', conductor.version)
        self.assertFalse(self.dbapi.check_versions())


class UpdateToLatestVersionsTestCase(base.DbTestCase):

    def setUp(self):
        super(UpdateToLatestVersionsTestCase, self).setUp()
        self.context = context.get_admin_context()
        self.dbapi = db_api.get_instance()

        obj_versions = release_mappings.get_object_versions(
            objects=['Node', 'Chassis'])
        master_objs = release_mappings.RELEASE_MAPPING['master']['objects']
        self.node_ver = master_objs['Node'][0]
        self.chassis_ver = master_objs['Chassis'][0]
        self.node_old_ver = self._get_old_object_version(
            self.node_ver, obj_versions['Node'])
        self.chassis_old_ver = self._get_old_object_version(
            self.chassis_ver, obj_versions['Chassis'])
        self.node_version_same = self.node_old_ver == self.node_ver
        self.chassis_version_same = self.chassis_old_ver == self.chassis_ver
        # number of objects with different versions
        self.num_diff_objs = 2
        if self.node_version_same:
            self.num_diff_objs -= 1
        if self.chassis_version_same:
            self.num_diff_objs -= 1

    def _get_old_object_version(self, latest_version, versions):
        """Return a version that is older (not same) as latest version.

        If there aren't any older versions, return the latest version.
        """
        for v in versions:
            if v != latest_version:
                return v
        return latest_version

    def test_empty_db(self):
        self.assertEqual(
            (0, 0), self.dbapi.update_to_latest_versions(self.context, 10))

    def test_version_exists(self):
        # Node will be in latest version
        utils.create_test_node()
        self.assertEqual(
            (0, 0), self.dbapi.update_to_latest_versions(self.context, 10))

    def test_one_node(self):
        node = utils.create_test_node(version=self.node_old_ver)
        expected = (0, 0) if self.node_version_same else (1, 1)
        self.assertEqual(
            expected, self.dbapi.update_to_latest_versions(self.context, 10))
        res = self.dbapi.get_node_by_uuid(node.uuid)
        self.assertEqual(self.node_ver, res.version)

    def test_max_count_zero(self):
        orig_node = utils.create_test_node(version=self.node_old_ver)
        orig_chassis = utils.create_test_chassis(version=self.chassis_old_ver)
        self.assertEqual((self.num_diff_objs, self.num_diff_objs),
                         self.dbapi.update_to_latest_versions(self.context, 0))
        node = self.dbapi.get_node_by_uuid(orig_node.uuid)
        self.assertEqual(self.node_ver, node.version)
        chassis = self.dbapi.get_chassis_by_uuid(orig_chassis.uuid)
        self.assertEqual(self.chassis_ver, chassis.version)

    def test_old_version_max_count_1(self):
        orig_node = utils.create_test_node(version=self.node_old_ver)
        orig_chassis = utils.create_test_chassis(version=self.chassis_old_ver)
        num_modified = 1 if self.num_diff_objs else 0
        self.assertEqual((self.num_diff_objs, num_modified),
                         self.dbapi.update_to_latest_versions(self.context, 1))
        node = self.dbapi.get_node_by_uuid(orig_node.uuid)
        chassis = self.dbapi.get_chassis_by_uuid(orig_chassis.uuid)
        self.assertTrue(node.version == self.node_old_ver
                        or chassis.version == self.chassis_old_ver)
        self.assertTrue(node.version == self.node_ver
                        or chassis.version == self.chassis_ver)

    def _create_nodes(self, num_nodes):
        version = self.node_old_ver
        nodes = []
        for i in range(0, num_nodes):
            node = utils.create_test_node(version=version,
                                          uuid=uuidutils.generate_uuid())
            # Create entries on the tables so we force field upgrades
            utils.create_test_node_trait(node_id=node.id, trait='foo',
                                         version='0.0')
            utils.create_test_bios_setting(node_id=node.id, version='1.0')

            nodes.append(node.uuid)
        for uuid in nodes:
            node = self.dbapi.get_node_by_uuid(uuid)
            self.assertEqual(version, node.version)
        return nodes

    def test_old_version_max_count_2_some_nodes(self):
        if self.node_version_same:
            # can't test if we don't have diff versions of the node
            return

        nodes = self._create_nodes(5)
        # Check/migrate 2, 10 remain.
        self.assertEqual(
            (10, 2), self.dbapi.update_to_latest_versions(self.context, 2))
        # Check/migrate 10, 8 migrated, 8 remain.
        self.assertEqual(
            (8, 8), self.dbapi.update_to_latest_versions(self.context, 10))
        # Just make sure it is still 0, 0 in case more things are added.
        self.assertEqual(
            (0, 0), self.dbapi.update_to_latest_versions(self.context, 10))
        for uuid in nodes:
            node = self.dbapi.get_node_by_uuid(uuid)
            self.assertEqual(self.node_ver, node.version)

    def test_old_version_max_count_same_nodes(self):
        if self.node_version_same:
            # can't test if we don't have diff versions of the node
            return
        vm_count = 5
        nodes = self._create_nodes(vm_count)
        # NOTE(TheJulia): Under current testing, 5 node will result in 10
        # records implicitly needing to be migrated.
        migrate_count = vm_count * 2
        self.assertEqual(
            (migrate_count, migrate_count),
            self.dbapi.update_to_latest_versions(self.context,
                                                 migrate_count))
        self.assertEqual(
            (0, 0), self.dbapi.update_to_latest_versions(self.context,
                                                         migrate_count))

        for uuid in nodes:
            node = self.dbapi.get_node_by_uuid(uuid)
            self.assertEqual(self.node_ver, node.version)


class MigrateToBuiltinInspectionTestCase(base.DbTestCase):

    def setUp(self):
        super().setUp()
        self.context = context.get_admin_context()
        self.dbapi = db_api.get_instance()
        CONF.set_override('enabled_inspect_interfaces', 'agent,no-inspect')

        for _ in range(3):
            utils.create_test_node(uuid=uuidutils.generate_uuid(),
                                   inspect_interface='inspector')
        for _ in range(2):
            utils.create_test_node(uuid=uuidutils.generate_uuid(),
                                   inspect_interface='agent')
        utils.create_test_node(uuid=uuidutils.generate_uuid(),
                               inspect_interface='no-inspect')

    def _check(self, migrated, left):
        current = sorted(x[0] for x in self.dbapi.get_nodeinfo_list(
            columns=['inspect_interface']))
        self.assertEqual(['agent'] * migrated + ['inspector'] * left
                         + ['no-inspect'], current)

    def test_migrate_all(self):
        total, migrated = self.dbapi.migrate_to_builtin_inspection(
            self.context, 0)
        self.assertEqual(3, total)
        self.assertEqual(3, migrated)
        self._check(5, 0)

    def test_cannot_migrate(self):
        CONF.set_override('enabled_inspect_interfaces', 'inspector,no-inspect')
        total, migrated = self.dbapi.migrate_to_builtin_inspection(
            self.context, 0)
        self.assertEqual(0, total)
        self.assertEqual(0, migrated)
        self._check(2, 3)

    def test_cannot_migrate_some(self):
        for state in [states.INSPECTING, states.INSPECTWAIT,
                      states.INSPECTFAIL]:
            utils.create_test_node(uuid=uuidutils.generate_uuid(),
                                   inspect_interface='inspector',
                                   provision_state=state)
        total, migrated = self.dbapi.migrate_to_builtin_inspection(
            self.context, 0)
        self.assertEqual(6, total)
        self.assertEqual(3, migrated)
        self._check(5, 3)

    def test_migrate_with_limit(self):
        total, migrated = self.dbapi.migrate_to_builtin_inspection(
            self.context, 2)
        self.assertEqual(3, total)
        self.assertEqual(2, migrated)
        self._check(4, 1)



# Copyright 2019 Red Hat, Inc.
# All Rights Reserved.
# Copyright (c) 2019-2021 Dell Inc. or its subsidiaries.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
Test class for DRAC boot interface
"""

from unittest import mock

import sushy

from ironic.common import boot_devices
from ironic.conductor import task_manager
from ironic.drivers.modules import deploy_utils
from ironic.drivers.modules.redfish import utils as redfish_utils
from ironic.tests.unit.db import utils as db_utils
from ironic.tests.unit.drivers.modules.drac import utils as test_utils
from ironic.tests.unit.objects import utils as obj_utils

INFO_DICT = dict(db_utils.get_test_redfish_info(), **test_utils.INFO_DICT)


@mock.patch.object(redfish_utils, 'get_system', autospec=True)
class DracBootTestCase(test_utils.BaseDracTest):

    def setUp(self):
        super(DracBootTestCase, self).setUp()
        self.node = obj_utils.create_test_node(
            self.context, driver='idrac', driver_info=INFO_DICT)

    @mock.patch.object(deploy_utils, 'validate_image_properties',
                       autospec=True)
    def test_validate_correct_vendor(self, mock_get_system,
                                     mock_validate_image_properties):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            task.node.instance_info.update(
                {'kernel': 'kernel',
                 'ramdisk': 'ramdisk',
                 'image_source': 'http://image/source'}
            )

            task.node.driver_info.update(
                {'deploy_kernel': 'kernel',
                 'deploy_ramdisk': 'ramdisk',
                 'bootloader': 'bootloader'}
            )

            task.node.properties['vendor'] = "Dell Inc."

            task.driver.boot.validate(task)

    def test__set_boot_device_persistent(self, mock_get_system):
        mock_manager = mock.MagicMock()
        mock_system = mock_get_system.return_value
        mock_system.managers = [mock_manager]
        mock_manager_oem = mock_manager.get_oem_extension.return_value

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            task.driver.boot._set_boot_device(
                task, boot_devices.CDROM, persistent=True)

            mock_manager_oem.set_virtual_boot_device.assert_called_once_with(
                sushy.VIRTUAL_MEDIA_CD, persistent=True, system=mock_system)

    def test__set_boot_device_cd(self, mock_get_system):
        mock_system = mock_get_system.return_value
        mock_manager = mock.MagicMock()
        mock_system.managers = [mock_manager]
        mock_manager_oem = mock_manager.get_oem_extension.return_value

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            task.driver.boot._set_boot_device(task, boot_devices.CDROM)

            mock_manager_oem.set_virtual_boot_device.assert_called_once_with(
                sushy.VIRTUAL_MEDIA_CD, persistent=False, system=mock_system)

    def test__set_boot_device_floppy(self, mock_get_system):
        mock_system = mock_get_system.return_value
        mock_manager = mock.MagicMock()
        mock_system.managers = [mock_manager]
        mock_manager_oem = mock_manager.get_oem_extension.return_value

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            task.driver.boot._set_boot_device(task, boot_devices.FLOPPY)

            mock_manager_oem.set_virtual_boot_device.assert_called_once_with(
                sushy.VIRTUAL_MEDIA_FLOPPY, persistent=False,
                system=mock_system)

    def test__set_boot_device_disk(self, mock_get_system):
        mock_system = mock_get_system.return_value

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            task.driver.boot._set_boot_device(task, boot_devices.DISK)

            self.assertFalse(mock_system.called)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from ironic.conductor import task_manager
from ironic.conf import CONF
from ironic.drivers.modules.inspector.hooks import pci_devices as \
    pci_devices_hook
from ironic.tests.unit.db import base as db_base
from ironic.tests.unit.objects import utils as obj_utils

_PLUGIN_DATA = {
    'pci_devices': [
        {
            'vendor_id': '8086', 'product_id': '2922', 'class': '010601',
            'revision': '02', 'bus': '0000:00:1f.2'
        },
        {
            'vendor_id': '8086', 'product_id': '2918', 'class': '060100',
            'revision': '02', 'bus': '0000:00:1f.0'
        },
        {
            'vendor_id': '8086', 'product_id': '2930', 'class': '0c0500',
            'revision': '02', 'bus': '0000:00:1f.3'
        },
        {
            'vendor_id': '1b36', 'product_id': '000c', 'class': '060400',
            'revision': '00', 'bus': '0000:00:01.2'
        },
        {
            'vendor_id': '1b36', 'product_id': '000c', 'class': '060400',
            'revision': '00', 'bus': '0000:00:01.0'
        },
        {
            'vendor_id': '1b36', 'product_id': '000d', 'class': '0c0330',
            'revision': '01', 'bus': '0000:02:00.0'
        }
    ]
}

_ALIASES = {('8086', '2922'): 'EightyTwentyTwo',
            ('8086', '2918'): 'EightyEighteen',
            ('1b36', '000c'): 'OneBZeroC'}


class PciDevicesTestCase(db_base.DbTestCase):
    def setUp(self):
        super().setUp()
        CONF.set_override('enabled_inspect_interfaces',
                          ['agent', 'no-inspect'])
        self.node = obj_utils.create_test_node(self.context,
                                               inspect_interface='agent')
        self.inventory = {'fake': 'fake-inventory'}
        self.plugin_data = _PLUGIN_DATA
        self.pci_devices_hook = pci_devices_hook.PciDevicesHook()
        self.pci_devices_hook._aliases = _ALIASES

    def test_pci_devices(self):
        with task_manager.acquire(self.context, self.node.id) as task:
            self.pci_devices_hook.__call__(task, self.inventory,
                                           self.plugin_data)
            self.node.refresh()
            result = self.node.properties.get('capabilities', '')
            expected = 'EightyTwentyTwo:1,EightyEighteen:1,OneBZeroC:2'
            self.assertEqual(expected, result)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
In-band inspection implementation.
"""

from oslo_config import cfg
from oslo_log import log as logging

from ironic.common import boot_devices
from ironic.common import exception
from ironic.common.i18n import _
from ironic.common import states
from ironic.conductor import utils as cond_utils
from ironic.drivers.modules import deploy_utils
from ironic.drivers.modules import inspect_utils
from ironic.drivers.modules.inspector import interface as common
from ironic.drivers import utils as drivers_utils

LOG = logging.getLogger(__name__)

CONF = cfg.CONF


class AgentInspect(common.Common):
    """In-band inspection."""

    default_require_managed_boot = True

    def __init__(self):
        super().__init__()
        enabled_hooks = [x.strip()
                         for x in CONF.inspector.hooks.split(',')
                         if x.strip()]
        self.hooks = inspect_utils.validate_inspection_hooks("agent",
                                                             enabled_hooks)

    def _start_managed_inspection(self, task):
        """Start inspection with boot managed by ironic."""
        ep = deploy_utils.get_ironic_api_url().rstrip('/')
        if ep.endswith('/v1'):
            ep = f'{ep}/continue_inspection'
        else:
            ep = f'{ep}/v1/continue_inspection'

        common.prepare_managed_inspection(task, ep)
        self._power_on_or_reboot(task)

    def _start_unmanaged_inspection(self, task):
        """Start unmanaged inspection."""
        try:
            if not task.node.disable_power_off:
                cond_utils.node_power_action(task, states.POWER_OFF)
            # Only network boot is supported for unmanaged inspection.
            cond_utils.node_set_boot_device(task, boot_devices.PXE,
                                            persistent=False)
            self._power_on_or_reboot(task)
        except Exception as exc:
            LOG.exception('Unable to start unmanaged inspection for node '
                          '%(uuid)s: %(err)s',
                          {'uuid': task.node.uuid, 'err': exc})
            error = _('unable to start inspection: %s') % exc
            common.inspection_error_handler(task, error, raise_exc=True,
                                            clean_up=False)

    def abort(self, task):
        """Abort hardware inspection.

        :param task: a task from TaskManager.
        """
        if inspect_utils.clear_lookup_addresses(task.node):
            task.node.save()

        common.clean_up(task, finish=False, always_power_off=True)

    def continue_inspection(self, task, inventory, plugin_data):
        """Continue in-band hardware inspection.

        :param task: a task from TaskManager.
        :param inventory: hardware inventory from the node.
        :param plugin_data: optional plugin-specific data.
        """
        # Run the inspection hooks
        inspect_utils.run_inspection_hooks(task, inventory, plugin_data,
                                           self.hooks, _store_logs)
        if CONF.agent.deploy_logs_collect == 'always':
            _store_logs(plugin_data, task.node)
        common.clean_up(task, finish=False, always_power_off=True)


def _store_logs(plugin_data, node):
    logs = plugin_data.get('logs')
    if not logs:
        LOG.warning('No logs were passed by the ramdisk for node %(node)s.',
                    {'node': node.uuid})
        return

    try:
        drivers_utils.store_ramdisk_logs(node, logs, label='inspect')
    except exception:
        LOG.exception('Could not store the ramdisk logs for node %(node)s. ',
                      {'node': node.uuid})



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

from ironic.conductor import task_manager
from ironic.conf import CONF
from ironic.drivers.modules.inspector.hooks import extra_hardware as hook
from ironic.tests.unit.db import base as db_base
from ironic.tests.unit.objects import utils as obj_utils


_PLUGIN_DATA = {
    'data': [
        ['disk', 'logical', 'count', '1'],
        ['disk', 'vda', 'size', '11'],
        ['disk', 'vda', 'vendor', '0x1af4'],
        ['disk', 'vda', 'physical_block_size', '512'],
        ['disk', 'vda', 'rotational', '1'],
        ['system', 'product', 'name', 'RHEL'],
        ['system', 'product', 'vendor', 'Red Hat'],
        ['system', 'product', 'version', 'RHEL-9.2.0 PC'],
        ['system', 'product', 'uuid', 'afdd3896-de8d-4585-8214-627071e13552'],
        ['system', 'motherboard', 'name', 'RHEL'],
        ['system', 'motherboard', 'vendor', 'Red Hat'],
        ['system', 'motherboard', 'version', 'RHEL-9.2.0 PC']
    ]
}

_EXPECTED_PLUGIN_DATA = {
    'extra': {
        'disk': {
            'logical': {
                'count': 1
            },
            'vda': {
                'size': 11,
                'vendor': '0x1af4',
                'physical_block_size': 512,
                'rotational': 1,
            }
        },
        'system': {
            'product': {
                'name': 'RHEL',
                'vendor': 'Red Hat',
                'version': 'RHEL-9.2.0 PC',
                'uuid': 'afdd3896-de8d-4585-8214-627071e13552'
            },
            'motherboard': {
                'name': 'RHEL',
                'vendor': 'Red Hat',
                'version': 'RHEL-9.2.0 PC'
            }
        }
    }
}


@mock.patch.object(hook.LOG, 'warning', autospec=True)
class ExtraHardwareTestCase(db_base.DbTestCase):
    def setUp(self):
        super().setUp()
        CONF.set_override('enabled_inspect_interfaces',
                          ['agent', 'no-inspect'])
        self.node = obj_utils.create_test_node(self.context,
                                               inspect_interface='agent')
        self.inventory = {'inventory': 'fake-inventory'}
        self.plugin_data = _PLUGIN_DATA

    def test_valid_extra_hardware(self, mock_warn):
        with task_manager.acquire(self.context, self.node.id) as task:
            hook.ExtraHardwareHook().__call__(task, self.inventory,
                                              self.plugin_data)
            self.assertFalse(mock_warn.called)
            self.assertEqual(self.plugin_data, _EXPECTED_PLUGIN_DATA)

    def test_no_data_received(self, mock_warn):
        self.plugin_data = {'cats': 'meow'}
        with task_manager.acquire(self.context, self.node.id) as task:
            hook.ExtraHardwareHook().__call__(task, self.inventory,
                                              self.plugin_data)
            mock_warn.assert_called_once_with(
                'No extra hardware information was received from the ramdisk '
                'for node %s', task.node.uuid)
            self.assertEqual(self.plugin_data, {'cats': 'meow'})

    @mock.patch.object(hook.LOG, 'debug', autospec=True)
    def test_extra_hardware_with_errors(self, mock_debug, mock_warn):
        self.plugin_data = {'data':
                            [['memory', 'total', 'size', '4294967296'],
                             [],
                             ['cpu', 'physical', 'number', '1'],
                             ['cpu', 'physical', 'WUT'],
                             ['cpu', 'logical', 'number', '1']]
                            }
        with task_manager.acquire(self.context, self.node.id) as task:
            hook.ExtraHardwareHook().__call__(task, self.inventory,
                                              self.plugin_data)
            expected = {'extra': {
                'memory': {
                    'total': {
                        'size': 4294967296
                    }
                },
                'cpu': {
                    'physical': {
                        'number': 1
                    },
                    'logical': {
                        'number': 1
                    },
                }
            }}

            self.assertEqual(expected, self.plugin_data)
            # An empty list is not a warning, a bad record is.
            self.assertEqual(1, mock_warn.call_count)
            mock_debug.assert_called_once_with(
                'Deleting \"data\" key from plugin data of node %s as it is '
                'assumed unusable by inspection rules.', task.node.uuid)

    def test_invalid_data_strict_mode_off(self, mock_warn):
        invalid_plugin_data = {
            'data': [['memory', 'total', 'size', '4294967296'],
                     ['cpu', 'physical', 'number', '1'],
                     {'interface': 'eth1'}]}
        self.plugin_data = invalid_plugin_data
        with task_manager.acquire(self.context, self.node.id) as task:
            hook.ExtraHardwareHook().__call__(task, self.inventory,
                                              self.plugin_data)

            self.assertEqual(invalid_plugin_data, self.plugin_data)
            mock_warn.assert_called_once_with(
                'Extra hardware data was not in a recognised format, and will '
                'not be forwarded to inspection rules for node %s',
                task.node.uuid)

    @mock.patch.object(hook.LOG, 'debug', autospec=True)
    def test_invalid_data_strict_mode_on(self, mock_debug, mock_warn):
        CONF.set_override('extra_hardware_strict', True, group='inspector')
        self.plugin_data = {
            'data': [['memory', 'total', 'size', '4294967296'],
                     ['cpu', 'physical', 'WUT']]
        }
        with task_manager.acquire(self.context, self.node.id) as task:
            hook.ExtraHardwareHook().__call__(task, self.inventory,
                                              self.plugin_data)
            self.assertEqual({}, self.plugin_data)
            mock_warn.assert_called_once_with(
                'Extra hardware data was not in a recognised format, and will '
                'not be forwarded to inspection rules for node %s',
                task.node.uuid)
            mock_debug.assert_called_once_with(
                'Deleting \"data\" key from plugin data of node %s as it is '
                'malformed and strict mode is on.', task.node.uuid)



#   Copyright 2025 Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime

from oslo_config import cfg
from oslo_utils import timeutils

from ironic.common import exception
from ironic.common import vnc as vnc_utils
from ironic.tests.unit.db import base as db_base
from ironic.tests.unit.objects import utils as obj_utils

CONF = cfg.CONF


class VncTestCase(db_base.DbTestCase):

    def setUp(self):
        super(VncTestCase, self).setUp()
        self.node = obj_utils.create_test_node(self.context)

    def test_novnc_authorize(self):
        then = timeutils.utcnow()

        vnc_utils.novnc_authorize(self.node)

        # assert token and timestamp were created
        self.assertIsNotNone(
            self.node.driver_internal_info['novnc_secret_token'])
        created = self.node.driver_internal_info['novnc_secret_token_created']
        self.assertIsNotNone(created)
        created_dt = datetime.datetime.strptime(created,
                                                '%Y-%m-%dT%H:%M:%S.%f')
        self.assertLessEqual(then, created_dt)

    def test_novnc_unauthorize(self):
        vnc_utils.novnc_authorize(self.node)
        self.assertIn('novnc_secret_token', self.node.driver_internal_info)
        self.assertIn('novnc_secret_token_created',
                      self.node.driver_internal_info)

        vnc_utils.novnc_unauthorize(self.node)

        # assert token and timestamp were removed
        self.assertNotIn('novnc_secret_token', self.node.driver_internal_info)
        self.assertNotIn('novnc_secret_token_created',
                         self.node.driver_internal_info)

    def test_novnc_validate(self):

        vnc_utils.novnc_authorize(self.node)
        token = self.node.driver_internal_info['novnc_secret_token']

        # assert successful validate
        self.assertIsNone(vnc_utils.novnc_validate(self.node, token))

        # assert wrong token
        self.assertRaises(exception.NotAuthorized, vnc_utils.novnc_validate,
                          self.node, 'wrong token')

        vnc_utils.novnc_unauthorize(self.node)
        # assert unauthorized
        self.assertRaises(exception.NotAuthorized, vnc_utils.novnc_validate,
                          self.node, token)

    def test_novnc_validate_expired(self):

        vnc_utils.novnc_authorize(self.node)
        token = self.node.driver_internal_info['novnc_secret_token']

        # assert successful validate
        self.assertIsNone(vnc_utils.novnc_validate(self.node, token))

        # set the created date to one second before the timeout
        now = timeutils.utcnow()
        timeout = CONF.vnc.token_timeout
        time_delta = datetime.timedelta(seconds=timeout + 1)
        then = now - time_delta
        self.node.set_driver_internal_info('novnc_secret_token_created',
                                           then.isoformat())

        # assert expired token
        self.assertRaises(exception.NotAuthorized, vnc_utils.novnc_validate,
                          self.node, token)

    def test_token_valid_until(self):
        now = timeutils.utcnow()
        self.node.set_driver_internal_info('novnc_secret_token_created',
                                           now.isoformat())
        timeout = CONF.vnc.token_timeout
        time_delta = datetime.timedelta(seconds=timeout)

        # assert that the valid until date is exactly timeout after now
        self.assertEqual(now + time_delta,
                         vnc_utils.token_valid_until(self.node))

    def test_get_console(self):
        self.node.set_driver_internal_info('novnc_secret_token', 'asdf')
        uuid = self.node.uuid
        CONF.set_override('public_url', 'http://192.0.2.1:6090/vnc_auto.html',
                          group='vnc')

        # assert expected console URL
        self.assertEqual({
            'type': 'vnc',
            'url': 'http://192.0.2.1:6090/vnc_auto.html?'
                   f'path=websockify%3Fnode%3D{uuid}%26token%3Dasdf'
        }, vnc_utils.get_console(self.node))



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from alembic import op
import sqlalchemy as sa


"""add parent_node field

Revision ID: fe222f476baf
Revises: 4dbec778866e
Create Date: 2023-04-10 11:59:29.633401

"""

# revision identifiers, used by Alembic.
revision = 'fe222f476baf'
down_revision = '4dbec778866e'


def upgrade():
    op.add_column('nodes', sa.Column('parent_node', sa.String(length=36),
                                     nullable=True))
    op.create_index(
        'parent_node_idx', 'nodes', ['parent_node'], unique=False)



# -*- encoding: utf-8 -*-
#
# Copyright 2013 Hewlett-Packard Development Company, L.P.
# Copyright 2014 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
"""
Utility for caching master images.
"""

import os
import shutil
import tempfile
import threading
import time
import uuid

from oslo_concurrency import lockutils
from oslo_log import log as logging
from oslo_utils import fileutils

from ironic.common import exception
from ironic.common.glance_service import service_utils
from ironic.common.i18n import _
from ironic.common import image_service
from ironic.common import images
from ironic.common import utils
from ironic.conf import CONF


LOG = logging.getLogger(__name__)

# This would contain a sorted list of instances of ImageCache to be
# considered for cleanup. This list will be kept sorted in non-increasing
# order of priority.
_cache_cleanup_list = []


_concurrency_semaphore = threading.Semaphore(CONF.image_download_concurrency)


class ImageCache(object):
    """Class handling access to cache for master images."""

    def __init__(self, master_dir, cache_size, cache_ttl,
                 disable_validation=False, force_raw=True):
        """Constructor.

        :param master_dir: cache directory to work on
                           Value of None disables image caching.
        :param cache_size: desired maximum cache size in bytes
        :param cache_ttl: cache entity TTL in seconds
        :param disable_validation: disable security checks on cached images
        """
        self.master_dir = master_dir
        self._cache_size = cache_size
        self._cache_ttl = cache_ttl
        self._force_raw = force_raw
        self._disable_validation = disable_validation
        if master_dir is not None:
            fileutils.ensure_tree(master_dir)

    def fetch_image(self, href, dest_path, ctx=None, force_raw=None,
                    expected_format=None, expected_checksum=None,
                    expected_checksum_algo=None,
                    image_auth_data=None):
        """Fetch image by given href to the destination path.

        Does nothing if destination path exists and is up to date with cache
        and href contents.
        Only creates a hard link (dest_path) to cached image if requested
        image is already in cache and up to date with href contents.
        Otherwise downloads an image, stores it in cache and creates a hard
        link (dest_path) to it.

        :param href: image UUID or href to fetch
        :param dest_path: destination file path
        :param ctx: context
        :param force_raw: boolean value, whether to convert the image to raw
                          format
        :param expected_format: The expected image format.
        :param expected_checksum: The expected image checksum
        :param expected_checksum_algo: The expected image checksum algorithm,
                                       if needed/supplied.
        """
        force_raw = force_raw if force_raw is not None else self._force_raw
        if expected_format is not None and self._disable_validation:
            raise AssertionError("BUG: passing expected_format to caches with "
                                 "disabled validation makes no sense")
        img_download_lock_name = 'download-image'
        if self.master_dir is None:
            # NOTE(ghe): We don't share images between instances/hosts
            # NOTE(TheJulia): These is a weird code path, because master_dir
            # has to be None, which by default it never should be unless
            # an operator forces it to None, which is a path we just never
            # expect.
            # TODO(TheJulia): This may be dead-ish code and likely needs
            # to be removed. Likely originated *out* of just the iscsi
            # deployment interface and local image caching.
            if not CONF.parallel_image_downloads:
                with lockutils.lock(img_download_lock_name):
                    _fetch(ctx, href, dest_path, force_raw,
                           expected_format=expected_format,
                           expected_checksum=expected_checksum,
                           expected_checksum_algo=expected_checksum_algo,
                           disable_validation=self._disable_validation,
                           image_auth_data=image_auth_data)
            else:
                with _concurrency_semaphore:
                    _fetch(ctx, href, dest_path, force_raw,
                           expected_format=expected_format,
                           expected_checksum=expected_checksum,
                           expected_checksum_algo=expected_checksum_algo,
                           disable_validation=self._disable_validation,
                           image_auth_data=image_auth_data)
            return

        # TODO(ghe): have hard links and counts the same behaviour in all fs

        # NOTE(vdrok): File name is converted to UUID if it's not UUID already,
        # so that two images with same file names do not collide
        if service_utils.is_glance_image(href):
            master_file_name = service_utils.parse_image_id(href)
        else:
            master_file_name = str(uuid.uuid5(uuid.NAMESPACE_URL, href))
        # NOTE(kaifeng) The ".converted" suffix acts as an indicator that the
        # image cached has gone through the conversion logic.
        if force_raw:
            master_file_name = master_file_name + '.converted'

        master_path = os.path.join(self.master_dir, master_file_name)

        if CONF.parallel_image_downloads:
            img_download_lock_name = 'download-image:%s' % master_file_name

        # TODO(dtantsur): lock expiration time
        with lockutils.lock(img_download_lock_name):
            img_service = image_service.get_image_service(href, context=ctx)
            if img_service.is_auth_set_needed:
                # We need to possibly authenticate based on what a user
                # has supplied, so we'll send that along.
                img_service.set_image_auth(href, image_auth_data)
            img_info = img_service.show(href)
            # NOTE(vdrok): After rebuild requested image can change, so we
            # should ensure that dest_path and master_path (if exists) are
            # pointing to the same file and their content is up to date
            cache_up_to_date = _delete_master_path_if_stale(master_path, href,
                                                            img_info)
            dest_up_to_date = _delete_dest_path_if_stale(master_path,
                                                         dest_path)

            if cache_up_to_date and dest_up_to_date:
                LOG.debug("Destination %(dest)s already exists "
                          "for image %(href)s",
                          {'href': href, 'dest': dest_path})
                return

            if cache_up_to_date:
                # NOTE(dtantsur): ensure we're not in the middle of clean up
                with lockutils.lock('master_image'):
                    try:
                        os.link(master_path, dest_path)
                    except OSError as exc:
                        LOG.debug(
                            "Could not hardlink image file %(image)s to "
                            "the cache location %(dest_path)s (will copy it "
                            "over): %(error)s", {
                                'image': master_path,
                                'dest_path': dest_path,
                                'error': exc})
                        shutil.copyfile(master_path, dest_path)
                LOG.debug("Master cache hit for image %(href)s",
                          {'href': href})
                return

            LOG.info("Master cache miss for image %(href)s, will download",
                     {'href': href})
            self._download_image(
                href, master_path, dest_path, img_info,
                ctx=ctx, force_raw=force_raw,
                expected_format=expected_format,
                expected_checksum=expected_checksum,
                expected_checksum_algo=expected_checksum_algo,
                image_auth_data=image_auth_data)

        # NOTE(dtantsur): we increased cache size - time to clean up
        self.clean_up()

    def _download_image(self, href, master_path, dest_path, img_info,
                        ctx=None, force_raw=None, expected_format=None,
                        expected_checksum=None, expected_checksum_algo=None,
                        image_auth_data=None):
        """Download image by href and store at a given path.

        This method should be called with uuid-specific lock taken.

        :param href: image UUID or href to fetch
        :param master_path: destination master path
        :param dest_path: destination file path
        :param img_info: image information from the image service
        :param ctx: context
        :param force_raw: boolean value, whether to convert the image to raw
                          format
        :param expected_format: The expected original format for the image.
        :param expected_checksum: The expected image checksum.
        :param expected_checksum_algo: The expected image checksum algorithm.
        :param image_auth_data: Dictionary with credential details which may be
                                required to download the file.
        :raise ImageDownloadFailed: when the image cache and the image HTTP or
                                    TFTP location are on different file system,
                                    causing hard link to fail.
        """
        # TODO(ghe): timeout and retry for downloads
        # TODO(ghe): logging when image cannot be created
        tmp_dir = tempfile.mkdtemp(dir=self.master_dir)
        tmp_path = os.path.join(tmp_dir, os.path.basename(master_path))
        force_raw = force_raw if force_raw is not None else self._force_raw
        try:
            with _concurrency_semaphore:
                _fetch(ctx, href, tmp_path, force_raw, expected_format,
                       expected_checksum=expected_checksum,
                       expected_checksum_algo=expected_checksum_algo,
                       disable_validation=self._disable_validation,
                       image_auth_data=image_auth_data)

            if img_info.get('no_cache'):
                LOG.debug("Caching is disabled for image %s", href)
                # Cache disabled, link directly to destination
                os.link(tmp_path, dest_path)
            else:
                # NOTE(dtantsur): no need for global lock here - master_path
                # will have link count >1 at any moment, so won't be cleaned up
                os.link(tmp_path, master_path)
                os.link(master_path, dest_path)
        except OSError as exc:
            msg = (_("Could not link image %(img_href)s from %(src_path)s "
                     "to %(dst_path)s, error: %(exc)s") %
                   {'img_href': href, 'src_path': master_path,
                    'dst_path': dest_path, 'exc': exc})
            LOG.error(msg)
            raise exception.ImageDownloadFailed(msg)
        finally:
            utils.rmtree_without_raise(tmp_dir)

    @lockutils.synchronized('master_image')
    def clean_up(self, amount=None):
        """Clean up directory with images, keeping cache of the latest images.

        Files with link count >1 are never deleted.
        Protected by global lock, so that no one messes with master images
        after we get listing and before we actually delete files.

        :param amount: if present, amount of space to reclaim in bytes,
                       cleaning will stop, if this goal was reached,
                       even if it is possible to clean up more files
        """
        if self.master_dir is None:
            return

        LOG.debug("Starting clean up for master image cache %(dir)s",
                  {'dir': self.master_dir})

        amount_copy = amount
        listing = _find_candidates_for_deletion(self.master_dir)
        survived, amount = self._clean_up_too_old(listing, amount)
        if amount is not None and amount <= 0:
            return
        amount = self._clean_up_ensure_cache_size(survived, amount)
        if amount is not None and amount > 0:
            LOG.warning("Cache clean up was unable to reclaim %(required)d "
                        "MiB of disk space, still %(left)d MiB required",
                        {'required': amount_copy / 1024 / 1024,
                         'left': amount / 1024 / 1024})

    def _clean_up_too_old(self, listing, amount):
        """Clean up stage 1: drop images that are older than TTL.

        This method removes files all files older than TTL seconds
        unless 'amount' is non-None. If 'amount' is non-None,
        it starts removing files older than TTL seconds,
        oldest first, until the required 'amount' of space is reclaimed.

        :param listing: list of tuples (file name, last used time)
        :param amount: if not None, amount of space to reclaim in bytes,
                       cleaning will stop, if this goal was reached,
                       even if it is possible to clean up more files
        :returns: tuple (list of files left after clean up,
                         amount still to reclaim)
        """
        threshold = time.time() - self._cache_ttl
        survived = []
        count = 0
        for file_name, last_used, stat in listing:
            if last_used < threshold:
                try:
                    os.unlink(file_name)
                except EnvironmentError as exc:
                    LOG.warning("Unable to delete file %(name)s from "
                                "master image cache: %(exc)s",
                                {'name': file_name, 'exc': exc})
                else:
                    count += 1
                    if amount is not None:
                        amount -= stat.st_size
                        if amount <= 0:
                            amount = 0
                            break
            else:
                survived.append((file_name, last_used, stat))
        if count:
            LOG.debug('Removed %(count)d expired file(s) from %(dir)s',
                      {'count': count, 'dir': self.master_dir})
        return survived, amount

    def _clean_up_ensure_cache_size(self, listing, amount):
        """Clean up stage 2: try to ensure cache size < threshold.

        Try to delete the oldest files until conditions is satisfied
        or no more files are eligible for deletion.

        :param listing: list of tuples (file name, last used time)
        :param amount: amount of space to reclaim, if possible.
                       if amount is not None, it has higher priority than
                       cache size in settings
        :returns: amount of space still required after clean up
        """
        # NOTE(dtantsur): Sort listing to delete the oldest files first
        listing = sorted(listing,
                         key=lambda entry: entry[1],
                         reverse=True)
        total_listing = (os.path.join(self.master_dir, f)
                         for f in os.listdir(self.master_dir))
        total_size = sum(os.path.getsize(f)
                         for f in total_listing)
        count = 0
        while listing and (total_size > self._cache_size
                           or (amount is not None and amount > 0)):
            file_name, last_used, stat = listing.pop()
            try:
                os.unlink(file_name)
            except EnvironmentError as exc:
                LOG.warning("Unable to delete file %(name)s from "
                            "master image cache: %(exc)s",
                            {'name': file_name, 'exc': exc})
            else:
                total_size -= stat.st_size
                count += 1
                if amount is not None:
                    amount -= stat.st_size

        if total_size > self._cache_size:
            LOG.info("After cleaning up cache dir %(dir)s "
                     "cache size %(actual)d is still larger than "
                     "threshold %(expected)d",
                     {'dir': self.master_dir, 'actual': total_size,
                      'expected': self._cache_size})
        elif count:
            LOG.debug(
                'Removed %(count)d file(s) from %(dir)s to free up space',
                {'count': count, 'dir': self.master_dir})
        return max(amount, 0) if amount is not None else 0


def _find_candidates_for_deletion(master_dir):
    """Find files eligible for deletion i.e. with link count ==1.

    :param master_dir: directory to operate on
    :returns: iterator yielding tuples (file name, last used time, stat)
    """
    for filename in os.listdir(master_dir):
        filename = os.path.join(master_dir, filename)
        stat = os.stat(filename)
        if not os.path.isfile(filename) or stat.st_nlink > 1:
            continue
        # NOTE(dtantsur): Detect most recently accessed files,
        # seeing atime can be disabled by the mount option
        # Also include ctime as it changes when image is linked to
        last_used_time = max(stat.st_mtime, stat.st_atime, stat.st_ctime)
        yield filename, last_used_time, stat


def _free_disk_space_for(path):
    """Get free disk space on a drive where path is located."""
    stat = os.statvfs(path)
    return stat.f_frsize * stat.f_bavail


def _fetch(context, image_href, path, force_raw=False,
           expected_format=None, expected_checksum=None,
           expected_checksum_algo=None,
           disable_validation=False, image_auth_data=None):
    """Fetch image and convert to raw format if needed."""
    assert not (disable_validation and expected_format)
    path_tmp = "%s.part" % path
    if os.path.exists(path_tmp):
        LOG.warning("%s exist, assuming it's stale", path_tmp)
        os.remove(path_tmp)
    images.fetch(context, image_href, path_tmp, force_raw=False,
                 checksum=expected_checksum,
                 checksum_algo=expected_checksum_algo,
                 image_auth_data=image_auth_data)
    # By default, the image format is unknown
    image_format = None
    disable_dii = (disable_validation
                   or CONF.conductor.disable_deep_image_inspection)
    if not disable_dii:
        if not expected_format:
            # Call of last resort to check the image format. Caching other
            # artifacts like kernel/ramdisks are not going to have an expected
            # format known even if they are not passed to qemu-img.
            remote_image_format = images.image_show(
                context,
                image_href,
                image_auth_data=image_auth_data).get('disk_format')
        else:
            remote_image_format = expected_format
        image_format = images.safety_check_image(path_tmp)
        images.check_if_image_format_is_permitted(
            image_format, remote_image_format)

    # Notes(yjiang5): If glance can provide the virtual size information,
    # then we can firstly clean cache and then invoke images.fetch().
    if (force_raw
            and ((disable_dii
                 and images.force_raw_will_convert(image_href, path_tmp))
                 or (not disable_dii
                     and image_format not in images.RAW_IMAGE_FORMATS))):
        # NOTE(TheJulia): What is happening here is the rest of the logic
        # is hinged on force_raw, but also we don't need to take the entire
        # path *if* the image on disk is *already* raw. Depending on settings,
        # the path differs slightly because if we have deep image inspection,
        # we can just rely upon the inspection image format, otherwise we
        # need to ask the image format.

        required_space = images.converted_size(path_tmp, estimate=False)
        directory = os.path.dirname(path_tmp)
        try:
            _clean_up_caches(directory, required_space)
        except exception.InsufficientDiskSpace:

            # try again with an estimated raw size instead of the full size
            required_space = images.converted_size(path_tmp, estimate=True)
            try:
                _clean_up_caches(directory, required_space)
            except exception.InsufficientDiskSpace:
                LOG.error('Not enough space for estimated image size. '
                          'Consider lowering '
                          '[DEFAULT]raw_image_growth_factor=%s',
                          CONF.raw_image_growth_factor)
                raise
        images.image_to_raw(image_href, path, path_tmp)
    else:
        os.rename(path_tmp, path)


def _clean_up_caches(directory, amount):
    """Explicitly cleanup caches based on their priority (if required).

    :param directory: the directory (of the cache) to be freed up.
    :param amount: amount of space to reclaim.
    :raises: InsufficientDiskSpace exception, if we cannot free up enough space
             after trying all the caches.
    """
    free = _free_disk_space_for(directory)

    if amount < free:
        return

    # NOTE(dtantsur): filter caches, whose directory is on the same device
    st_dev = os.stat(directory).st_dev

    caches_to_clean = [x[1]() for x in _cache_cleanup_list]
    caches = (c for c in caches_to_clean
              if os.stat(c.master_dir).st_dev == st_dev)
    for cache_to_clean in caches:
        cache_to_clean.clean_up(amount=(amount - free))
        free = _free_disk_space_for(directory)
        if amount < free:
            break
    else:
        raise exception.InsufficientDiskSpace(path=directory,
                                              required=amount / 1024 / 1024,
                                              actual=free / 1024 / 1024,
                                              )


def clean_up_caches(ctx, directory, images_info, image_auth_data=None):
    """Explicitly cleanup caches based on their priority (if required).

    This cleans up the caches to free up the amount of space required for the
    images in images_info. The caches are cleaned up one after the other in
    the order of their priority.  If we still cannot free up enough space
    after trying all the caches, this method throws exception.

    :param ctx: context
    :param directory: the directory (of the cache) to be freed up.
    :param images_info: a list of tuples of the form (image_uuid,path)
                        for which space is to be created in cache.
    :raises: InsufficientDiskSpace exception, if we cannot free up enough space
             after trying all the caches.
    """
    total_size = sum(images.download_size(ctx, uuid,
                                          image_auth_data=image_auth_data)
                     for (uuid, path) in images_info)
    _clean_up_caches(directory, total_size)


def clean_up_all():
    """Clean up all entries from all caches."""
    caches_to_clean = [x[1]() for x in _cache_cleanup_list]
    for cache in caches_to_clean:
        cache.clean_up()


def cleanup(priority):
    """Decorator method for adding cleanup priority to a class."""
    def _add_property_to_class_func(cls):
        _cache_cleanup_list.append((priority, cls))
        _cache_cleanup_list.sort(reverse=True, key=lambda tuple_: tuple_[0])
        return cls

    return _add_property_to_class_func


def _delete_master_path_if_stale(master_path, href, img_info):
    """Delete image from cache if it is not up to date with href contents.

    :param master_path: path to an image in master cache
    :param href: image href
    :param img_info: image information from the service
    :returns: True if master_path is up to date with href contents,
        False if master_path was stale and was deleted or it didn't exist
    """
    if service_utils.is_glance_image(href):
        # Glance image contents cannot be updated without changing image's UUID
        return os.path.exists(master_path)
    if image_service.is_container_registry_url(href):
        # OCI Images cannot be changed without changing the digest values.
        return os.path.exists(master_path)
    if os.path.exists(master_path):
        img_mtime = img_info.get('updated_at')
        if not img_mtime:
            # This means that href is not a glance image and doesn't have an
            # updated_at attribute. To play on the safe side, redownload the
            # master copy of the image.
            LOG.warning("Image service couldn't determine last "
                        "modification time of %(href)s, updating "
                        "the cached copy %(cached_file)s.",
                        {'href': href, 'cached_file': master_path})
        else:
            master_mtime = utils.unix_file_modification_datetime(master_path)
            if img_mtime <= master_mtime:
                return True
            # Delete image from cache as it is outdated
            LOG.info('Image %(href)s was last modified at %(remote_time)s. '
                     'Deleting the cached copy "%(cached_file)s since it was '
                     'last modified at %(local_time)s and may be outdated.',
                     {'href': href, 'remote_time': img_mtime,
                      'local_time': master_mtime, 'cached_file': master_path})

        os.unlink(master_path)
    return False


def _delete_dest_path_if_stale(master_path, dest_path):
    """Delete dest_path if it does not point to cached image.

    :param master_path: path to an image in master cache
    :param dest_path: hard link to an image
    :returns: True if dest_path points to master_path, False if dest_path was
        stale and was deleted or it didn't exist
    """
    dest_path_exists = os.path.exists(dest_path)
    if not dest_path_exists:
        # Image not cached, re-download
        return False
    master_path_exists = os.path.exists(master_path)
    if (not master_path_exists
            or os.stat(master_path).st_ino != os.stat(dest_path).st_ino):
        # Image exists in cache, but dest_path out of date
        os.unlink(dest_path)
        return False
    return True



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


"""add node inventory table

Revision ID: 0ac0f39bc5aa
Revises: 9ef41f07cb58
Create Date: 2022-10-25 17:15:38.181544

"""

from alembic import op
from oslo_db.sqlalchemy import types as db_types
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '0ac0f39bc5aa'
down_revision = '9ef41f07cb58'


def upgrade():
    op.create_table('node_inventory',
                    sa.Column('version', sa.String(length=15), nullable=True),
                    sa.Column('created_at', sa.DateTime(), nullable=True),
                    sa.Column('updated_at', sa.DateTime(), nullable=True),
                    sa.Column('id', sa.Integer(), nullable=False),
                    sa.Column('inventory_data', db_types.JsonEncodedDict(
                        mysql_as_long=True).impl, nullable=True),
                    sa.Column('plugin_data', db_types.JsonEncodedDict(
                        mysql_as_long=True).impl, nullable=True),
                    sa.Column('node_id', sa.Integer(), nullable=True),
                    sa.PrimaryKeyConstraint('id'),
                    sa.ForeignKeyConstraint(['node_id'], ['nodes.id'], ),
                    sa.Index('inventory_node_id_idx', 'node_id'),
                    mysql_engine='InnoDB',
                    mysql_charset='UTF8MB3')



# Copyright (c) 2016 Red Hat, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

from oslo_config import cfg

from ironic.common import exception
from ironic.console.rfb import auth
from ironic.console.rfb import authnone
from ironic.console.rfb import auths
from ironic.tests import base

CONF = cfg.CONF


class RFBAuthSchemeListTestCase(base.TestCase):

    def setUp(self):
        super(RFBAuthSchemeListTestCase, self).setUp()

    def test_load_ok(self):
        schemelist = auths.RFBAuthSchemeList()

        security_types = sorted(schemelist.schemes.keys())
        self.assertEqual([auth.AuthType.NONE],
                         security_types)

    def test_load_unknown(self):
        """Ensure invalid auth schemes are not supported.

        We're really testing oslo_policy functionality, but this case is
        esoteric enough to warrant this.
        """
        self.assertRaises(
            ValueError, CONF.set_override, 'novnc_auth_schemes',
            ['none', 'wibble'], group='vnc')

    def test_find_scheme_ok(self):
        schemelist = auths.RFBAuthSchemeList()

        scheme = schemelist.find_scheme(
            [auth.AuthType.TIGHT,
             auth.AuthType.NONE])

        self.assertIsInstance(scheme, authnone.RFBAuthSchemeNone)

    def test_find_scheme_fail(self):
        schemelist = auths.RFBAuthSchemeList()

        self.assertRaises(exception.RFBAuthNoAvailableScheme,
                          schemelist.find_scheme,
                          [auth.AuthType.TIGHT])

    def test_find_scheme_priority(self):
        schemelist = auths.RFBAuthSchemeList()

        tight = mock.MagicMock(spec=auth.RFBAuthScheme)
        schemelist.schemes[auth.AuthType.TIGHT] = tight

        scheme = schemelist.find_scheme(
            [auth.AuthType.TIGHT,
             auth.AuthType.NONE])

        self.assertEqual(tight, scheme)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
"""
Base PXE Interface Methods
"""

from oslo_config import cfg
from oslo_log import log as logging

from ironic.common import boot_devices
from ironic.common import dhcp_factory
from ironic.common import exception
from ironic.common.i18n import _
from ironic.common import metrics_utils
from ironic.common import pxe_utils
from ironic.common import states
from ironic.conductor import periodics
from ironic.conductor import utils as manager_utils
from ironic.drivers.modules import boot_mode_utils
from ironic.drivers.modules import deploy_utils
from ironic.drivers import utils as driver_utils


CONF = cfg.CONF

LOG = logging.getLogger(__name__)

METRICS = metrics_utils.get_metrics_logger(__name__)

REQUIRED_PROPERTIES = {
    'deploy_kernel': _("UUID (from Glance) of the deployment kernel. "
                       "Required."),
    'deploy_ramdisk': _("UUID (from Glance) of the ramdisk that is "
                        "mounted at boot time. Required."),
}
RESCUE_PROPERTIES = {
    'rescue_kernel': _('UUID (from Glance) of the rescue kernel. This value '
                       'is required for rescue mode.'),
    'rescue_ramdisk': _('UUID (from Glance) of the rescue ramdisk with agent '
                        'that is used at node rescue time. This value is '
                        'required for rescue mode.'),
}
OPTIONAL_PROPERTIES = {
    'kernel_append_params': driver_utils.KERNEL_APPEND_PARAMS_DESCRIPTION %
    {'option_group': 'pxe'},
}
COMMON_PROPERTIES = REQUIRED_PROPERTIES.copy()
COMMON_PROPERTIES.update(driver_utils.OPTIONAL_PROPERTIES)
COMMON_PROPERTIES.update(RESCUE_PROPERTIES)
COMMON_PROPERTIES.update(OPTIONAL_PROPERTIES)


class PXEBaseMixin(object):

    ipxe_enabled = False

    http_boot_enabled = False

    def get_properties(self):
        """Return the properties of the interface.

        :returns: dictionary of <property name>:<property description> entries.
        """
        return COMMON_PROPERTIES

    def _use_http_folder(self):
        if self.ipxe_enabled:
            return True
        if self.http_boot_enabled:
            return True
        return False

    @METRICS.timer('PXEBaseMixin.clean_up_ramdisk')
    def clean_up_ramdisk(self, task):
        """Cleans up the boot of ironic ramdisk.

        This method cleans up the PXE environment that was setup for booting
        the deploy or rescue ramdisk. It unlinks the deploy/rescue
        kernel/ramdisk in the node's directory in tftproot and removes it's PXE
        config.

        :param task: a task from TaskManager.
        :param mode: Label indicating a deploy or rescue operation
            was carried out on the node. Supported values are 'deploy' and
            'rescue'. Defaults to 'deploy', indicating deploy operation was
            carried out.
        :returns: None
        """
        node = task.node
        mode = deploy_utils.rescue_or_deploy_mode(node)
        try:
            images_info = pxe_utils.get_image_info(
                node, mode=mode, ipxe_enabled=self._use_http_folder())
        except exception.MissingParameterValue as e:
            LOG.warning('Could not get %(mode)s image info '
                        'to clean up images for node %(node)s: %(err)s',
                        {'mode': mode, 'node': node.uuid, 'err': e})
        else:
            pxe_utils.clean_up_pxe_env(
                task, images_info, ipxe_enabled=self._use_http_folder())

    @METRICS.timer('PXEBaseMixin.clean_up_instance')
    def clean_up_instance(self, task):
        """Cleans up the boot of instance.

        This method cleans up the environment that was setup for booting
        the instance. It unlinks the instance kernel/ramdisk in node's
        directory in tftproot and removes the PXE config.

        :param task: a task from TaskManager.
        :returns: None
        """
        node = task.node

        try:
            images_info = pxe_utils.get_instance_image_info(
                task, ipxe_enabled=self._use_http_folder())
        except exception.MissingParameterValue as e:
            LOG.warning('Could not get instance image info '
                        'to clean up images for node %(node)s: %(err)s',
                        {'node': node.uuid, 'err': e})
        else:
            pxe_utils.clean_up_pxe_env(task, images_info,
                                       ipxe_enabled=self._use_http_folder())

        boot_mode_utils.deconfigure_secure_boot_if_needed(task)

    @METRICS.timer('PXEBaseMixin.prepare_ramdisk')
    def prepare_ramdisk(self, task, ramdisk_params):
        """Prepares the boot of Ironic ramdisk using PXE.

        This method prepares the boot of the deploy or rescue kernel/ramdisk
        after reading relevant information from the node's driver_info and
        instance_info.

        :param task: a task from TaskManager.
        :param ramdisk_params: the parameters to be passed to the ramdisk.
            pxe driver passes these parameters as kernel command-line
            arguments.
        :returns: None
        :raises: MissingParameterValue, if some information is missing in
            node's driver_info or instance_info.
        :raises: InvalidParameterValue, if some information provided is
            invalid.
        :raises: IronicException, if some power or set boot boot device
            operation failed on the node.
        """
        node = task.node
        # Label indicating a deploy or rescue operation being carried out on
        # the node, 'deploy' or 'rescue'. Unless the node is in a rescue like
        # state, the mode is set to 'deploy', indicating deploy operation is
        # being carried out.
        mode = deploy_utils.rescue_or_deploy_mode(node)

        if self.ipxe_enabled:
            # NOTE(mjturek): At this point, the ipxe boot script should
            # already exist as it is created at startup time. However, we
            # call the boot script create method here to assert its
            # existence and handle the unlikely case that it wasn't created
            # or was deleted.
            pxe_utils.create_ipxe_boot_script()

        # Generate options for both IPv4 and IPv6, and they can be
        # filtered down later based upon the port options.
        # TODO(TheJulia): This should be re-tooled during the Victoria
        # development cycle so that we call a single method and return
        # combined options. The method we currently call is relied upon
        # by two eternal projects, to changing the behavior is not ideal.
        dhcp_opts = pxe_utils.dhcp_options_for_instance(
            task, ipxe_enabled=self.ipxe_enabled, ip_version=4,
            http_boot_enabled=self.http_boot_enabled)
        dhcp_opts += pxe_utils.dhcp_options_for_instance(
            task, ipxe_enabled=self.ipxe_enabled, ip_version=6,
            http_boot_enabled=self.http_boot_enabled)
        provider = dhcp_factory.DHCPFactory()
        provider.update_dhcp(task, dhcp_opts)
        # TODO(TheJulia): We need to change the parameter name for
        # ipxe_enabled in pxe_utils at some point since here it is
        # an indicator of where to put the files on the filesystem.
        pxe_info = pxe_utils.get_image_info(
            node, mode=mode,
            ipxe_enabled=self._use_http_folder())

        # NODE: Try to validate and fetch instance images only
        # if we are in DEPLOYING state.
        if node.provision_state == states.DEPLOYING:
            pxe_info.update(
                pxe_utils.get_instance_image_info(
                    task, ipxe_enabled=self.ipxe_enabled))

        boot_mode_utils.sync_boot_mode(task)

        pxe_options = pxe_utils.build_pxe_config_options(
            task, pxe_info, ipxe_enabled=self.ipxe_enabled,
            ramdisk_params=ramdisk_params)
        # TODO(dtantsur): backwards compatibility hack, remove in the V release
        if ramdisk_params.get("ipa-api-url"):
            pxe_options["ipa-api-url"] = ramdisk_params["ipa-api-url"]

        if self.ipxe_enabled:
            pxe_config_template = deploy_utils.get_ipxe_config_template(node)
        else:
            pxe_config_template = deploy_utils.get_pxe_config_template(node)

        pxe_utils.create_pxe_config(task, pxe_options,
                                    pxe_config_template,
                                    ipxe_enabled=self.ipxe_enabled)
        self._node_set_boot_device_for_network_boot(task)

        if self.ipxe_enabled and CONF.pxe.ipxe_use_swift:
            kernel_label = '%s_kernel' % mode
            ramdisk_label = '%s_ramdisk' % mode
            pxe_info.pop(kernel_label, None)
            pxe_info.pop(ramdisk_label, None)

        if pxe_info:
            pxe_utils.cache_ramdisk_kernel(
                task, pxe_info,
                ipxe_enabled=self._use_http_folder())

        LOG.debug('Ramdisk (i)PXE boot for node %(node)s has been prepared '
                  'with kernel params %(params)s',
                  {'node': node.uuid, 'params': pxe_options})

    def _node_set_boot_device_for_network_boot(self, task, persistent=False):
        """Helper to handle httpboot aware network booting.

        Basic challenge: IPMI doesn't have a field reserved for "httpboot" as
        httpboot pre-dates IPMI. It is also entirely possible that all logic
        to support httpboot is coming from OPROM code on network cards, so to
        sort of handle this, and the nature of PXE being perceived as
        "network boot", if we are http boot enabled, we attempt to explicitly
        request as such, but if the driver errors, then we fall back to PXE.

        :params task: a TaskManager object.
        :params persistent: Default False, if the network boot request is
                            persistent.
        """
        if self.http_boot_enabled:
            try:
                manager_utils.node_set_boot_device(task,
                                                   boot_devices.UEFIHTTP,
                                                   persistent=persistent)
            except exception.InvalidParameterValue:
                LOG.warning('Attempted to set HTTPBOOT for node %s, but it is '
                            'not supported by the driver. Falling back to '
                            'PXE to trigger network boot.', task.node.uuid)
                manager_utils.node_set_boot_device(task, boot_devices.PXE,
                                                   persistent=persistent)
        else:
            manager_utils.node_set_boot_device(task, boot_devices.PXE,
                                               persistent=persistent)

    @METRICS.timer('PXEBaseMixin.prepare_instance')
    def prepare_instance(self, task):
        """Prepares the boot of instance.

        This method prepares the boot of the instance after reading
        relevant information from the node's instance_info. In case of netboot,
        it updates the dhcp entries and switches the PXE config. In case of
        localboot, it cleans up the PXE config.

        :param task: a task from TaskManager.
        :returns: None
        """
        boot_mode_utils.sync_boot_mode(task)
        node = task.node
        boot_device = None
        boot_option = deploy_utils.get_boot_option(node)
        if boot_option != "kickstart":
            boot_mode_utils.configure_secure_boot_if_needed(task)

        instance_image_info = {}
        if boot_option == "ramdisk" or boot_option == "kickstart":
            instance_image_info = pxe_utils.get_instance_image_info(
                task, ipxe_enabled=self.ipxe_enabled)
            pxe_utils.cache_ramdisk_kernel(
                task, instance_image_info,
                ipxe_enabled=self._use_http_folder())
            if 'ks_template' in instance_image_info:
                ks_cfg = pxe_utils.validate_kickstart_template(
                    instance_image_info['ks_template'][1]
                )
                pxe_utils.validate_kickstart_file(ks_cfg)

        if (deploy_utils.is_iscsi_boot(task) or boot_option == "ramdisk"
                or boot_option == "kickstart"):
            pxe_utils.prepare_instance_pxe_config(
                task, instance_image_info,
                iscsi_boot=deploy_utils.is_iscsi_boot(task),
                ramdisk_boot=(boot_option == "ramdisk"),
                anaconda_boot=(boot_option == "kickstart"),
                ipxe_enabled=self.ipxe_enabled,
                http_boot_enabled=self.http_boot_enabled)
            pxe_utils.prepare_instance_kickstart_config(
                task, instance_image_info,
                anaconda_boot=(boot_option == "kickstart"))
            boot_device = boot_devices.PXE

        else:
            # NOTE(dtantsur): create a PXE configuration as a safety net for
            # hardware uncapable of persistent boot. If on a reboot it will try
            # to boot from PXE, this configuration will return it back.
            if CONF.pxe.enable_netboot_fallback:
                pxe_utils.build_service_pxe_config(
                    task, instance_image_info,
                    task.node.driver_internal_info.get('root_uuid_or_disk_id'),
                    ipxe_enabled=self.ipxe_enabled,
                    # PXE config for whole disk images is identical to what
                    # we need to boot from local disk, so use True even
                    # for partition images.
                    is_whole_disk_image=True)
            else:
                # Clean up the deployment configuration
                pxe_utils.clean_up_pxe_config(
                    task, ipxe_enabled=self.ipxe_enabled)
            boot_device = boot_devices.DISK

        # NOTE(pas-ha) do not re-set boot device on ACTIVE nodes
        # during takeover
        if boot_device and (task.node.provision_state not in
                            (states.ACTIVE, states.ADOPTING)):
            if boot_device == boot_devices.PXE:
                # Implying network booting, we need to handle the case
                # it might be HTTPBoot instead of PXEBoot.
                self._node_set_boot_device_for_network_boot(task,
                                                            persistent=True)
            else:
                vendor = task.node.properties.get('vendor', None)
                boot_mode = boot_mode_utils.get_boot_mode(task.node)
                if (task.node.provision_state == states.DEPLOYING
                        and vendor and vendor.lower() == 'lenovo'
                        and boot_mode == 'uefi'
                        and boot_device == boot_devices.DISK):
                    # Lenovo hardware is modeled on a "just update"
                    # UEFI nvram model of use, and if multiple actions
                    # get requested, you can end up in cases where NVRAM
                    # changes are deleted as the host "restores" to the
                    # backup. For more information see
                    # https://bugs.launchpad.net/ironic/+bug/2053064
                    return
                manager_utils.node_set_boot_device(task, boot_device,
                                                   persistent=True)

    def _validate_common(self, task):
        node = task.node

        if not driver_utils.get_node_mac_addresses(task):
            raise exception.MissingParameterValue(
                _("Node %s does not have any port associated with it")
                % node.uuid)

        if self.ipxe_enabled:
            if not CONF.deploy.http_url or not CONF.deploy.http_root:
                raise exception.MissingParameterValue(_(
                    "iPXE boot is enabled but no HTTP URL or HTTP "
                    "root was specified"))

        # NOTE(zer0c00l): When 'kickstart' boot option is used we need to store
        # kickstart and squashfs files in http_root directory. These files
        # will be eventually requested by anaconda installer during deployment
        # over http(s).
        if deploy_utils.get_boot_option(node) == 'kickstart':
            if not CONF.deploy.http_url or not CONF.deploy.http_root:
                raise exception.MissingParameterValue(_(
                    "'kickstart' boot option is set on the node but no HTTP "
                    "URL or HTTP root was specified"))

            if not CONF.anaconda.default_ks_template:
                raise exception.MissingParameterValue(_(
                    "'kickstart' boot option is set on the node but no "
                    "default kickstart template is specified"))

        deploy_utils.validate_capabilities(node)

        # Check if we have invalid parameters being passed which will not work
        # for ramdisk configurations.
        if (node.instance_info.get('image_source')
            and node.instance_info.get('boot_iso')):
            raise exception.InvalidParameterValue(_(
                "An 'image_source' and 'boot_iso' parameter may not be "
                "specified at the same time"))

        pxe_utils.parse_driver_info(node)

    @METRICS.timer('PXEBaseMixin.validate')
    def validate(self, task):
        """Validate the PXE-specific info for booting deploy/instance images.

        This method validates the PXE-specific info for booting the
        ramdisk and instance on the node.  If invalid, raises an
        exception; otherwise returns None.

        :param task: a task from TaskManager.
        :returns: None
        :raises: InvalidParameterValue, if some parameters are invalid.
        :raises: MissingParameterValue, if some required parameters are
            missing.
        """
        self._validate_common(task)
        node = task.node

        # NOTE(TheJulia): If we're not writing an image, we can skip
        # the remainder of this method.
        # NOTE(dtantsur): if we're are writing an image with local boot
        # the boot interface does not care about image parameters and
        # must not validate them.
        boot_option = deploy_utils.get_boot_option(node)
        if (not task.driver.storage.should_write_image(task)
                or boot_option == 'local'):
            return

        d_info = deploy_utils.get_image_instance_info(node)
        deploy_utils.validate_image_properties(task, d_info)

    @METRICS.timer('PXEBaseMixin.validate_rescue')
    def validate_rescue(self, task):
        """Validate that the node has required properties for rescue.

        :param task: a TaskManager instance with the node being checked
        :raises: MissingParameterValue if node is missing one or more required
            parameters
        """
        pxe_utils.parse_driver_info(task.node, mode='rescue')

    @METRICS.timer('PXEBaseMixin.validate_inspection')
    def validate_inspection(self, task):
        """Validate that the node has required properties for inspection.

        :param task: A TaskManager instance with the node being checked
        :raises: UnsupportedDriverExtension
        """
        try:
            self._validate_common(task)
        except exception.MissingParameterValue as exc:
            # Fall back to non-managed in-band inspection
            raise exception.UnsupportedDriverExtension(
                _("Insufficient information provided for managed "
                  "inspection: %s") % exc)

    _RETRY_ALLOWED_STATES = {states.DEPLOYWAIT, states.CLEANWAIT,
                             states.RESCUEWAIT}

    @METRICS.timer('PXEBaseMixin._check_boot_timeouts')
    @periodics.node_periodic(
        purpose='checking PXE boot status',
        spacing=CONF.pxe.boot_retry_check_interval,
        enabled=bool(CONF.pxe.boot_retry_timeout),
        filters={'provision_state_in': _RETRY_ALLOWED_STATES,
                 'reserved': False,
                 'maintenance': False,
                 'provisioned_before': CONF.pxe.boot_retry_timeout},
    )
    def _check_boot_timeouts(self, task, manager, context):
        """Periodically checks whether boot has timed out and retry it.

        :param task: a task instance.
        :param manager: conductor manager.
        :param context: request context.
        """
        self._check_boot_status(task)

    def _check_boot_status(self, task):
        if not _should_retry_boot(task.node):
            return

        task.upgrade_lock(purpose='retrying PXE boot')

        # Retry critical checks after acquiring the exclusive lock.
        if (task.node.maintenance or task.node.provision_state
                not in self._RETRY_ALLOWED_STATES
                or not _should_retry_boot(task.node)):
            return

        LOG.info('Booting the ramdisk on node %(node)s is taking more than '
                 '%(timeout)d seconds, retrying boot',
                 {'node': task.node.uuid,
                  'timeout': CONF.pxe.boot_retry_timeout})

        manager_utils.node_power_action(task, states.POWER_OFF)
        self._node_set_boot_device_for_network_boot(task)
        manager_utils.node_power_action(task, states.POWER_ON)


def _should_retry_boot(node):
    # NOTE(dtantsur): this assumes IPA, do we need to make it generic?
    for field in ('agent_last_heartbeat', 'last_power_state_change'):
        if node.driver_internal_info.get('agent_secret_token', False):
            LOG.debug('Not retrying PXE boot for node %(node)s; an agent '
                      'token has been identified, meaning the agent '
                      'has started.',
                      {'node': node.uuid})
            return False
        if manager_utils.value_within_timeout(
                node.driver_internal_info.get(field),
                CONF.pxe.boot_retry_timeout):
            # Alive and heartbeating, probably busy with something long
            LOG.debug('Not retrying PXE boot for node %(node)s; its '
                      '%(event)s happened less than %(timeout)d seconds ago',
                      {'node': node.uuid, 'event': field,
                       'timeout': CONF.pxe.boot_retry_timeout})
            return False
    return True



#
# Copyright 2022 Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg

from ironic.common.i18n import _

opts = [
    cfg.StrOpt('power_delay',
               default='0',
               help=_('Delay in seconds for operations with the fake '
                      'power driver. Two comma-delimited values will '
                      'result in a delay with a triangular random '
                      'distribution, weighted on the first value.')),
    cfg.StrOpt('boot_delay',
               default='0',
               help=_('Delay in seconds for operations with the fake '
                      'boot driver. Two comma-delimited values will '
                      'result in a delay with a triangular random '
                      'distribution, weighted on the first value.')),
    cfg.StrOpt('deploy_delay',
               default='0',
               help=_('Delay in seconds for operations with the fake '
                      'deploy driver. Two comma-delimited values will '
                      'result in a delay with a triangular random '
                      'distribution, weighted on the first value.')),
    cfg.StrOpt('vendor_delay',
               default='0',
               help=_('Delay in seconds for operations with the fake '
                      'vendor driver. Two comma-delimited values will '
                      'result in a delay with a triangular random '
                      'distribution, weighted on the first value.')),
    cfg.StrOpt('management_delay',
               default='0',
               help=_('Delay in seconds for operations with the fake '
                      'management driver. Two comma-delimited values will '
                      'result in a delay with a triangular random '
                      'distribution, weighted on the first value.')),
    cfg.StrOpt('inspect_delay',
               default='0',
               help=_('Delay in seconds for operations with the fake '
                      'inspect driver. Two comma-delimited values will '
                      'result in a delay with a triangular random '
                      'distribution, weighted on the first value.')),
    cfg.StrOpt('raid_delay',
               default='0',
               help=_('Delay in seconds for operations with the fake '
                      'raid driver. Two comma-delimited values will '
                      'result in a delay with a triangular random '
                      'distribution, weighted on the first value.')),
    cfg.StrOpt('bios_delay',
               default='0',
               help=_('Delay in seconds for operations with the fake '
                      'bios driver. Two comma-delimited values will '
                      'result in a delay with a triangular random '
                      'distribution, weighted on the first value.')),
    cfg.StrOpt('storage_delay',
               default='0',
               help=_('Delay in seconds for operations with the fake '
                      'storage driver. Two comma-delimited values will '
                      'result in a delay with a triangular random '
                      'distribution, weighted on the first value.')),
    cfg.StrOpt('rescue_delay',
               default='0',
               help=_('Delay in seconds for operations with the fake '
                      'rescue driver. Two comma-delimited values will '
                      'result in a delay with a triangular random '
                      'distribution, weighted on the first value.')),
    cfg.StrOpt('firmware_delay',
               default='0',
               help=_('Delay in seconds for operations with the fake '
                      'firmware driver. Two comma-delimited values will '
                      'result in a delay with a triangular random '
                      'distribution, weighted on the first value.')),
]


def register_opts(conf):
    conf.register_opts(opts, group='fake')



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


"""Fault definitions."""

POWER_FAILURE = 'power failure'
""" Node is moved to maintenance due to power synchronization failure. """

CLEAN_FAILURE = 'clean failure'
""" Node is moved to maintenance due to failure of a cleaning
    operation. """

RESCUE_ABORT_FAILURE = 'rescue abort failure'
""" Node is moved to maintenance due to failure of cleaning up during
    rescue abort. """

SERVICE_FAILURE = 'service failure'
""" Node is moved to maintenance due to failure of a service operation. """

VALID_FAULTS = (POWER_FAILURE, CLEAN_FAILURE, RESCUE_ABORT_FAILURE,
                SERVICE_FAILURE)



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Tests for nova websocketproxy."""

import fixtures
import io
import socket
from unittest import mock

from oslo_config import cfg
import oslo_middleware.cors as cors_middleware
from oslo_utils import timeutils

from ironic.common import exception
from ironic.common import vnc as vnc_utils

from ironic.console.securityproxy import base
from ironic.console import websocketproxy
from ironic.tests.unit.db import base as db_base
from ironic.tests.unit.objects import utils as obj_utils

CONF = cfg.CONF


class IronicProxyRequestHandlerDBTestCase(db_base.DbTestCase):

    def setUp(self):
        super(IronicProxyRequestHandlerDBTestCase, self).setUp()

        self.node = obj_utils.create_test_node(
            self.context,
            driver_internal_info={
                'vnc_host': 'node1',
                'vnc_port': 10000,
                'novnc_secret_token': '123-456-789',
                'novnc_secret_token_created': timeutils.utcnow().isoformat()
            }
        )
        self.uuid = self.node.uuid

        with mock.patch('websockify.ProxyRequestHandler', autospec=True):
            self.wh = websocketproxy.IronicProxyRequestHandler()
        self.wh.server = websocketproxy.IronicWebSocketProxy()
        self.wh.socket = mock.MagicMock()
        self.wh.msg = mock.MagicMock()
        self.wh.do_proxy = mock.MagicMock()
        self.wh.headers = mock.MagicMock()
        # register [cors] config options
        cors_middleware.CORS(None, CONF)
        CONF.set_override(
            'allowed_origin',
            ['allowed-origin-example-1.net', 'allowed-origin-example-2.net'],
            group='cors')

    fake_header = {
        'cookie': 'token="123-456-789"',
        'Origin': 'https://example.net:6080',
        'Host': 'example.net:6080',
    }

    @mock.patch('ironic.common.vnc.novnc_validate',
                autospec=True)
    @mock.patch('ironic.objects.Node.get_by_uuid',
                autospec=True)
    def test_new_websocket_client_db(
            self, mock_node_get, mock_validate,
            node_not_found=False):

        if node_not_found:
            mock_node_get.side_effect = exception.NodeNotFound(
                node=self.uuid)
        else:
            mock_node_get.return_value = self.node

        tsock = mock.MagicMock()
        tsock.recv.return_value = "HTTP/1.1 200 OK\r\n\r\n"
        self.wh.socket.return_value = tsock

        self.wh.path = "http://127.0.0.1/?token=123-456-789"
        self.wh.headers = self.fake_header

        if node_not_found:
            self.assertRaises(exception.NotAuthorized,
                              self.wh.new_websocket_client)
        else:
            self.wh.new_websocket_client()
            mock_validate.assert_called_once_with(mock.ANY, '123-456-789')
            self.wh.socket.assert_called_with('node1', 10000, connect=True)
            self.wh.do_proxy.assert_called_with(tsock)

    def test_new_websocket_client_db_instance_not_found(self):
        self.test_new_websocket_client_db(node_not_found=True)


class IronicProxyRequestHandlerTestCase(db_base.DbTestCase):

    def setUp(self):
        super(IronicProxyRequestHandlerTestCase, self).setUp()

        self.node = obj_utils.create_test_node(
            self.context,
            driver_internal_info={
                'vnc_host': 'node1',
                'vnc_port': 10000,
                'novnc_secret_token': '123-456-789',
                'novnc_secret_token_created': timeutils.utcnow().isoformat()
            }
        )
        self.uuid = self.node.uuid

        self.server = websocketproxy.IronicWebSocketProxy()
        with mock.patch('websockify.ProxyRequestHandler', autospec=True):
            self.wh = websocketproxy.IronicProxyRequestHandler()
        self.wh.server = self.server
        self.wh.socket = mock.MagicMock()
        self.wh.msg = mock.MagicMock()
        self.wh.do_proxy = mock.MagicMock()
        self.wh.headers = mock.MagicMock()
        # register [cors] config options
        cors_middleware.CORS(None, CONF)
        CONF.set_override(
            'allowed_origin',
            ['allowed-origin-example-1.net', 'allowed-origin-example-2.net'],
            group='cors')

        self.threading_timer_mock = self.useFixture(
            fixtures.MockPatch('threading.Timer', mock.DEFAULT)).mock

    fake_header = {
        'cookie': 'token="123-456-789"',
        'Origin': 'https://example.net:6080',
        'Host': 'example.net:6080',
    }

    fake_header_ipv6 = {
        'cookie': 'token="123-456-789"',
        'Origin': 'https://[2001:db8::1]:6080',
        'Host': '[2001:db8::1]:6080',
    }

    fake_header_bad_token = {
        'cookie': 'token="XXX"',
        'Origin': 'https://example.net:6080',
        'Host': 'example.net:6080',
    }

    fake_header_bad_origin = {
        'cookie': 'token="123-456-789"',
        'Origin': 'https://bad-origin-example.net:6080',
        'Host': 'example.net:6080',
    }

    fake_header_allowed_origin = {
        'cookie': 'token="123-456-789"',
        'Origin': 'https://allowed-origin-example-2.net:6080',
        'Host': 'example.net:6080',
    }

    fake_header_blank_origin = {
        'cookie': 'token="123-456-789"',
        'Origin': '',
        'Host': 'example.net:6080',
    }

    fake_header_no_origin = {
        'cookie': 'token="123-456-789"',
        'Host': 'example.net:6080',
    }

    fake_header_http = {
        'cookie': 'token="123-456-789"',
        'Origin': 'http://example.net:6080',
        'Host': 'example.net:6080',
    }

    fake_header_malformed_cookie = {
        'cookie': '?=!; token="123-456-789"',
        'Origin': 'https://example.net:6080',
        'Host': 'example.net:6080',
    }

    @mock.patch('ironic.common.vnc.novnc_validate',
                autospec=True)
    def test_new_websocket_client(self, validate):
        validate.return_value = '123-456-789'

        self.wh.socket.return_value = '<socket>'
        self.wh.path = f"http://127.0.0.1/?node={self.uuid}&token=123-456-789"
        self.wh.headers = self.fake_header

        self.wh.new_websocket_client()

        validate.assert_called_with(mock.ANY, '123-456-789')
        self.wh.socket.assert_called_with('node1', 10000, connect=True)
        self.wh.do_proxy.assert_called_with('<socket>')

    @mock.patch('ironic.common.vnc.novnc_validate',
                autospec=True)
    def test_new_websocket_client_ipv6_url(self, validate):
        validate.return_value = '123-456-789'

        tsock = mock.MagicMock()
        self.wh.socket.return_value = tsock
        ip = '[2001:db8::1]'
        self.wh.path = f"http://{ip}/?node={self.uuid}&token=123-456-789"
        self.wh.headers = self.fake_header_ipv6

        self.wh.new_websocket_client()

        validate.assert_called_with(mock.ANY, "123-456-789")
        self.wh.socket.assert_called_with('node1', 10000, connect=True)
        self.wh.do_proxy.assert_called_with(tsock)

    @mock.patch('ironic.common.vnc.novnc_validate',
                autospec=True)
    def test_new_websocket_client_token_invalid(self, validate):
        validate.side_effect = exception.NotAuthorized()

        self.wh.path = f"http://127.0.0.1/?node={self.uuid}&token=XXX"
        self.wh.headers = self.fake_header_bad_token

        self.assertRaises(exception.NotAuthorized,
                          self.wh.new_websocket_client)
        validate.assert_called_with(mock.ANY, "XXX")

    @mock.patch('socket.getfqdn',
                autospec=True)
    def test_address_string_doesnt_do_reverse_dns_lookup(self, getfqdn):
        request_mock = mock.MagicMock()
        request_mock.makefile().readline.side_effect = [
            b'GET /vnc_auth.html?token=123-456-789 HTTP/1.1\r\n',
            b''
        ]
        server_mock = mock.MagicMock()
        client_address = ('8.8.8.8', 54321)

        handler = websocketproxy.IronicProxyRequestHandler(
            request_mock, client_address, server_mock)
        handler.log_message('log message using client address context info')

        self.assertFalse(getfqdn.called)  # no reverse dns look up
        self.assertEqual(handler.address_string(), '8.8.8.8')  # plain address

    @mock.patch('ironic.common.vnc.novnc_validate',
                autospec=True)
    def test_new_websocket_client_novnc_bad_origin_header(self, validate):
        validate.return_value = '123-456-789'

        self.wh.path = f"http://127.0.0.1/?node={self.uuid}&token=123-456-789"
        self.wh.headers = self.fake_header_bad_origin

        self.assertRaises(exception.NotAuthorized,
                          self.wh.new_websocket_client)

    @mock.patch('ironic.common.vnc.novnc_validate',
                autospec=True)
    def test_new_websocket_client_novnc_allowed_origin_header(self, validate):
        validate.return_value = '123-456-789'

        tsock = mock.MagicMock()
        self.wh.socket.return_value = tsock
        self.wh.path = f"http://127.0.0.1/?node={self.uuid}&token=123-456-789"
        self.wh.headers = self.fake_header_allowed_origin

        self.wh.new_websocket_client()

        validate.assert_called_with(mock.ANY, "123-456-789")
        self.wh.socket.assert_called_with('node1', 10000, connect=True)
        self.wh.do_proxy.assert_called_with(tsock)

    @mock.patch('ironic.common.vnc.novnc_validate',
                autospec=True)
    def test_new_websocket_client_novnc_blank_origin_header(self, validate):
        validate.return_value = '123-456-789'

        self.wh.path = f"http://127.0.0.1/?node={self.uuid}&token=123-456-789"
        self.wh.headers = self.fake_header_blank_origin

        self.assertRaises(exception.NotAuthorized,
                          self.wh.new_websocket_client)

    @mock.patch('ironic.common.vnc.novnc_validate',
                autospec=True)
    def test_new_websocket_client_novnc_no_origin_header(self, validate):
        validate.return_value = '123-456-789'

        tsock = mock.MagicMock()
        self.wh.socket.return_value = tsock

        self.wh.path = f"http://127.0.0.1/?node={self.uuid}&token=123-456-789"
        self.wh.headers = self.fake_header_no_origin

        self.wh.new_websocket_client()

        self.wh.socket.assert_called_with('node1', 10000, connect=True)
        self.wh.do_proxy.assert_called_with(tsock)

    @mock.patch('ironic.common.vnc.novnc_validate',
                autospec=True)
    def test_new_websocket_client_http_forwarded_proto_https(self, validate):
        validate.return_value = '123-456-789'

        header = {
            'cookie': 'token="123-456-789"',
            'Origin': 'http://example.net:6080',
            'Host': 'example.net:6080',
            'X-Forwarded-Proto': 'https'
        }
        self.wh.socket.return_value = '<socket>'
        self.wh.path = f"https://127.0.0.1/?node={self.uuid}&token=123-456-789"
        self.wh.headers = header

        self.wh.new_websocket_client()

        validate.assert_called_with(mock.ANY, "123-456-789")
        self.wh.socket.assert_called_with('node1', 10000, connect=True)
        self.wh.do_proxy.assert_called_with('<socket>')

    def test_reject_open_redirect(self, url='//example.com/%2F..'):
        # This will test the behavior when an attempt is made to cause an open
        # redirect. It should be rejected.
        mock_req = mock.MagicMock()
        mock_req.makefile().readline.side_effect = [
            f'GET {url} HTTP/1.1\r\n'.encode('utf-8'),
            b''
        ]

        client_addr = ('8.8.8.8', 54321)
        mock_server = mock.MagicMock()
        # This specifies that the server will be able to handle requests other
        # than only websockets.
        mock_server.only_upgrade = False

        # Constructing a handler will process the mock_req request passed in.
        handler = websocketproxy.IronicProxyRequestHandler(
            mock_req, client_addr, mock_server)

        # Collect the response data to verify at the end. The
        # SimpleHTTPRequestHandler writes the response data to a 'wfile'
        # attribute.
        output = io.BytesIO()
        handler.wfile = output
        # Process the mock_req again to do the capture.
        handler.do_GET()
        output.seek(0)
        result = output.readlines()

        # Verify no redirect happens and instead a 400 Bad Request is returned.
        # NOTE: As of python 3.10.6 there is a fix for this vulnerability,
        # which will cause a 301 Moved Permanently error to be returned
        # instead that redirects to a sanitized version of the URL with extra
        # leading '/' characters removed.
        # See https://github.com/python/cpython/issues/87389 for details.
        # We will consider either response to be valid for this test. This will
        # also help if and when the above fix gets backported to older versions
        # of python.
        errmsg = result[0].decode()
        expected_ironic = '400 URI must not start with //'
        expected_cpython = '301 Moved Permanently'

        self.assertTrue(expected_ironic in errmsg
                        or expected_cpython in errmsg)

        # If we detect the cpython fix, verify that the redirect location is
        # now the same url but with extra leading '/' characters removed.
        if expected_cpython in errmsg:
            location = result[3].decode()
            if location.startswith('Location: '):
                location = location[len('Location: '):]
            location = location.rstrip('\r\n')
            self.assertTrue(
                location.startswith('/example.com/%2F..'),
                msg='Redirect location is not the expected sanitized URL',
            )

    def test_reject_open_redirect_3_slashes(self):
        self.test_reject_open_redirect(url='///example.com/%2F..')

    @mock.patch('websockify.websocketproxy.select_ssl_version',
                autospec=True)
    def test_ssl_min_version_is_not_set(self, mock_select_ssl):
        websocketproxy.IronicWebSocketProxy()
        self.assertFalse(mock_select_ssl.called)

    @mock.patch('websockify.websocketproxy.select_ssl_version',
                autospec=True)
    def test_ssl_min_version_not_set_by_default(self, mock_select_ssl):
        websocketproxy.IronicWebSocketProxy(ssl_minimum_version='default')
        self.assertFalse(mock_select_ssl.called)

    @mock.patch('websockify.websocketproxy.select_ssl_version',
                autospec=True)
    def test_non_default_ssl_min_version_is_set(self, mock_select_ssl):
        minver = 'tlsv1_3'
        websocketproxy.IronicWebSocketProxy(ssl_minimum_version=minver)
        mock_select_ssl.assert_called_once_with(minver)

    def test__close_connection(self):
        tsock = mock.MagicMock()
        self.wh.vmsg = mock.MagicMock()
        host = 'node1'
        port = '10000'

        self.wh._close_connection(tsock, host, port)
        tsock.shutdown.assert_called_once_with(socket.SHUT_RDWR)
        tsock.close.assert_called_once()
        self.wh.vmsg.assert_called_once_with(
            f"{host}:{port}: Websocket client or target closed")

    def test__close_connection_raise_OSError(self):
        tsock = mock.MagicMock()
        self.wh.vmsg = mock.MagicMock()
        host = 'node1'
        port = '10000'

        tsock.shutdown.side_effect = OSError("Error")

        self.wh._close_connection(tsock, host, port)

        tsock.shutdown.assert_called_once_with(socket.SHUT_RDWR)
        tsock.close.assert_called_once()

        self.wh.vmsg.assert_called_once_with(
            f"{host}:{port}: Websocket client or target closed")


class IronicWebsocketSecurityProxyTestCase(db_base.DbTestCase):

    def setUp(self):
        super(IronicWebsocketSecurityProxyTestCase, self).setUp()

        self.server = websocketproxy.IronicWebSocketProxy(
            security_proxy=mock.MagicMock(
                spec=base.SecurityProxy)
        )

        self.node = obj_utils.create_test_node(self.context)
        vnc_utils.novnc_authorize(self.node)
        uuid = self.node.uuid
        token = self.node.driver_internal_info['novnc_secret_token']

        with mock.patch('websockify.ProxyRequestHandler',
                        autospec=True):
            self.wh = websocketproxy.IronicProxyRequestHandler()
        self.wh.server = self.server
        self.wh.path = f"http://127.0.0.1/?node={uuid}&token={token}"
        self.wh.socket = mock.MagicMock()
        self.wh.msg = mock.MagicMock()
        self.wh.do_proxy = mock.MagicMock()
        self.wh.headers = mock.MagicMock()

        # register [cors] config options
        cors_middleware.CORS(None, CONF)

        def get_header(header):
            if header == 'Origin':
                return 'https://example.net:6080'
            elif header == 'Host':
                return 'example.net:6080'
            else:
                return

        self.wh.headers.get = get_header

    @mock.patch('ironic.console.websocketproxy.TenantSock.close',
                autospec=True)
    @mock.patch('ironic.console.websocketproxy.TenantSock.finish_up',
                autospec=True)
    def test_proxy_connect_ok(self, mock_finish, mock_close):

        sock = mock.MagicMock(
            spec=websocketproxy.TenantSock)
        self.server.security_proxy.connect.return_value = sock

        self.wh.new_websocket_client()

        self.wh.do_proxy.assert_called_with(sock)
        mock_finish.assert_called()
        mock_close.assert_not_called()

    @mock.patch('ironic.console.websocketproxy.TenantSock.close',
                autospec=True)
    @mock.patch('ironic.console.websocketproxy.TenantSock.finish_up',
                autospec=True)
    def test_proxy_connect_err(self, mock_finish, mock_close):

        ex = exception.SecurityProxyNegotiationFailed("Wibble")
        self.server.security_proxy.connect.side_effect = ex

        self.assertRaises(exception.SecurityProxyNegotiationFailed,
                          self.wh.new_websocket_client)

        self.assertEqual(len(self.wh.do_proxy.calls), 0)
        mock_close.assert_called()
        mock_finish.assert_not_called()
